from typing import Callable, List

import re
import typeguard
import valid8


validate = valid8.validate


@typeguard.typechecked
def pattern(regex: str) -> Callable[[str], bool]:
    r = re.compile(regex)

    def res(value):
        return bool(r.fullmatch(value))

    res.__name__ = f'pattern({regex})'
    return res


def does_not_has_attributes(cls, attributes: List[str], and_annotations: bool = False):
    if and_annotations:
        validate(
            f"{cls.__name__} has no attribute __annotations__",
            hasattr(cls, '__annotations__') and cls.__annotations__ != {},
            equals=False
        )
    for attribute in attributes:
        validate(f"{cls.__name__} has no attribute {attribute}", hasattr(cls, attribute), equals=False)



import dataclasses
from dataclasses import InitVar
from typing import List, Final, Optional

import typeguard
from selenium import webdriver
from selenium.common import WebDriverException, NoSuchElementException
from selenium.webdriver import Keys
from selenium.webdriver.common.by import By

from dumbo_esse3.primitives import Course, Username, Password, Exam, Student, StudentThesisState, CdL, \
    ExamDescription, ExamNotes, ExamType, DateTime, Register, NumberOfHours, Semester, RegisterActivity, \
    ActivityTitle, ActivityType, GraduationDay, StudentGraduation
from dumbo_esse3.utils import validators
from dumbo_esse3.utils.validators import validate

ESSE3_SERVER = "https://unical.esse3.cineca.it"
URLs: Final = {
    "login": f'{ESSE3_SERVER}/auth/Logon.do?menu_opened_cod=',
    "logout": f'{ESSE3_SERVER}/Logout.do?menu_opened_cod=',
    "course_list": f'{ESSE3_SERVER}/auth/docente/CalendarioEsami/ListaAttivitaCalEsa.do?menu_opened_cod=menu_link-navbox_docenti_Didattica',
    "thesis_list": f'{ESSE3_SERVER}/auth/docente/Graduation/LaureandiAssegnati.do?menu_opened_cod=menu_link-navbox_docenti_Conseguimento_Titolo',
    "register_list": f'{ESSE3_SERVER}/auth/docente/RegistroDocente/Home.do?menu_opened_cod=menu_link-navbox_docenti_Registro',
    "graduation_day_list": f'{ESSE3_SERVER}/auth/docente/Graduation/ElencoSeduteLaurea.do',
}


def change_esse3_server(url):
    global ESSE3_SERVER, URLs

    for key in URLs.keys():
        URLs[key] = URLs[key].replace(ESSE3_SERVER, url, 1)

    ESSE3_SERVER = url


@typeguard.typechecked
@dataclasses.dataclass(frozen=True)
class Esse3Wrapper:
    key: InitVar[object]
    username: InitVar[Username]
    password: InitVar[Password]
    debug: bool = dataclasses.field(default=False)
    driver: webdriver.Chrome = dataclasses.field(default_factory=webdriver.Chrome)
    __key = object()

    def __post_init__(self, key: object, username: Username, password: Password):
        validators.validate('key', key, equals=self.__key, help_msg="Can only be instantiated using a factory method")
        self.maximize()
        self.__login(username, password)

    def __del__(self):
        if not self.debug:
            try:
                self.__logout()
                self.driver.close()
            except WebDriverException:
                pass
            except ValueError:
                pass

    @classmethod
    def create(cls, username: str, password: str, debug: bool = False, detached: bool = False,
               headless: bool = True) -> 'Esse3Wrapper':
        options = webdriver.ChromeOptions()
        options.headless = headless
        if debug or detached:
            options.add_experimental_option("detach", True)
        driver = webdriver.Chrome(options=options)

        return Esse3Wrapper(
            key=cls.__key,
            username=Username.parse(username),
            password=Password.parse(password),
            debug=debug,
            driver=driver,
        )

    @property
    def is_headless(self) -> bool:
        return self.driver.execute_script("return navigator.webdriver")

    def __login(self, username: Username, password: Password) -> None:
        self.driver.get(URLs["login"])
        self.driver.find_element(By.ID, 'u').send_keys(username.value)
        self.driver.find_element(By.ID, 'p').send_keys(password.value)
        self.driver.find_element(By.ID, 'btnLogin').send_keys(Keys.RETURN)

    def __logout(self) -> None:
        self.driver.get(URLs["logout"])

    def minimize(self) -> None:
        self.driver.minimize_window()

    def maximize(self) -> None:
        self.driver.maximize_window()

    def fetch_courses(self) -> List[Course]:
        self.driver.get(URLs["course_list"])
        rows = self.driver.find_elements(By.XPATH, "//tr[@class='detail_table'][td//input[@src = 'images/sostenuta.gif']]")
        res = []
        for idx, row in enumerate(rows):
            td = row.find_element(By.XPATH, "td[1]")
            res.append(Course.of(td.text))
        return res

    def fetch_exams(self, course: Course) -> List[Exam]:
        self.driver.get(URLs["course_list"])
        self.driver.find_element(By.XPATH, f"//tr[td = '{course}']/td//input[@src = 'images/sostenuta.gif']").send_keys(Keys.RETURN)
        exams = self.driver.find_elements(By.XPATH, '//tr[@class="detail_table"]')
        return list(sorted([Exam.of(
            DateTime.parse(exam.find_element(By.XPATH, "td[3]").text),
            int(exam.find_element(By.XPATH, "td[5]").text or 0),
        ) for exam in exams], key=lambda exam: exam.date_and_time))

    def fetch_students(self, course: Course, exam: DateTime) -> List[Student]:
        self.driver.get(URLs["course_list"])
        self.driver.find_element(By.XPATH, f"//tr[td = '{course}']/td//input[@src = 'images/sostenuta.gif']").send_keys(Keys.RETURN)

        self.driver.find_element(By.XPATH, f"//tr[normalize-space(td/text()) = '{exam}']//input[@src='images/defAppStudent.gif']").send_keys(Keys.RETURN)

        rows = self.driver.find_elements(By.XPATH, f"//table[@class='detail_table']//tr[position() > 1][td[1] >= 1]")
        return [
            Student.of(
                row.find_element(By.XPATH, 'td[3]').text,
                row.find_element(By.XPATH, 'td[4]').text,
            ) for row in rows
        ]

    def is_exam_present(self, course: Course, date_and_time: DateTime) -> bool:
        exams = self.fetch_exams(course)
        return any(exam for exam in exams if exam.date_and_time == date_and_time)

    def add_exam(self, course: Course, exam: DateTime, exam_type: ExamType, description: ExamDescription,
                 notes: ExamNotes) -> None:
        self.driver.get(URLs["course_list"])
        self.driver.find_element(By.XPATH, f"//tr[td = '{course}']/td//input[@src = 'images/sostenuta.gif']").send_keys(Keys.RETURN)

        self.driver.find_element(By.XPATH, '//input[@type = "submit"][@name = "new_pf"]').send_keys(Keys.RETURN)

        self.driver.find_element(By.ID, 'data_inizio_app').send_keys(exam.stringify_date() + Keys.ESCAPE)
        self.driver.find_element(By.NAME, 'hh_esa').send_keys(exam.stringify_hour())
        self.driver.find_element(By.NAME, 'mm_esa').send_keys(exam.stringify_minute())

        self.driver.find_element(By.XPATH, f'//input[@type = "radio"][@value = "{exam_type.value}"]').send_keys(Keys.SPACE)

        self.driver.find_element(By.XPATH, '//tr[starts-with(th, "*Descrizione:")]//input').send_keys(description.value)
        self.driver.find_element(By.XPATH, '//tr[starts-with(th, "Note:")]//textarea').send_keys(notes.value)

        self.driver.find_element(By.NAME, 'sbmDef').send_keys(Keys.RETURN)
        self.driver.find_element(By.XPATH, '//a[. = "Esci"]').send_keys(Keys.RETURN)

    def fetch_thesis_list(self) -> List[StudentThesisState]:
        self.driver.get(URLs["thesis_list"])

        all_students = []

        tables = self.driver.find_elements(By.XPATH, f"//div[@id='containerPrincipale']/table")
        for table in tables:
            cdl = table.find_element(By.XPATH, f"caption").text
            students = table.find_elements(By.XPATH, f"tbody/tr")
            for student in students:
                all_students.append((
                    Student.of(
                        student.find_element(By.XPATH, f"td[1]").text,
                        student.find_element(By.XPATH, f"td[2]").text,
                    ),
                    CdL.parse(cdl),
                    student.find_element(By.XPATH, f"td[5]//a[@id = 'btnAllegatiTesi']").get_attribute("href")
                    if student.find_elements(By.XPATH, f"td[5]//a[@id = 'btnAllegatiTesi']") else None,
                ))

        res = []
        for student in all_students:
            state = StudentThesisState.State.MISSING
            if student[2]:
                state = StudentThesisState.State.UNSIGNED
                self.driver.get(student[2])
                if self.driver.find_elements(By.XPATH, f"//td[text() = 'Approvato']"):
                    state = StudentThesisState.State.SIGNED
                self.driver.back()
            res.append(StudentThesisState.of(student[0], student[1], state))
        return res

    def __thesis_action(self, student: Student, action) -> None:
        self.driver.get(URLs["thesis_list"])
        self.driver.find_element(By.XPATH, f"//tr[td/text() = '{student.id}'][td/text() = '{student.name}']//a[@id = 'btnAllegatiTesi']").send_keys(Keys.RETURN)
        self.driver.find_element(By.ID, action).send_keys(Keys.RETURN)

    def show_thesis(self, student: Student) -> None:
        self.__thesis_action(student, "btnDownload")

    def sign_thesis(self, student: Student) -> None:
        self.__thesis_action(student, "btnApprova")
        self.driver.find_element(By.ID, "selApprova1").send_keys(Keys.SPACE)
        self.driver.find_element(By.ID, "btnApprova").send_keys(Keys.RETURN)

    def fetch_registers(self) -> List[Register]:
        self.driver.get(URLs["register_list"])
        rows = self.driver.find_elements(By.XPATH, "//tr[td/a/img[@src = 'images/open_registro.gif']]")
        res = []
        for idx, row in enumerate(rows):
            course = Course.parse(row.find_element(By.XPATH, "td[2]").text)
            hours = NumberOfHours.parse(row.find_element(By.XPATH, "td[4]").text)
            semester = Semester.parse(row.find_element(By.XPATH, "td[5]").text)
            state = Register.State(row.find_element(By.XPATH, "td[6]/img").get_attribute('alt'))
            res.append(Register.of(course=course, hours=hours, semester=semester, state=state))
        return sorted(res, key=lambda register: (register.semester, register.course))

    def fetch_register_activities(self, register: Register, with_time: bool = False) -> List[RegisterActivity]:
        self.driver.get(URLs["register_list"])
        self.driver.find_element(
            By.XPATH,
            f"//tr[normalize-space(td/text()) = '{register.course}']/td//a[img/@src = 'images/open_registro.gif']"
        ).send_keys(Keys.RETURN)

        rows_xpath = "//tr[td/form/input/@src = 'images/open_registro.gif']"
        rows = self.driver.find_elements(By.XPATH, rows_xpath)
        number_of_rows = len(rows)
        res = []
        for index in range(number_of_rows):
            row = rows[index]
            date = DateTime.parse_date(row.find_element(By.XPATH, "td[2]").text)
            hours = NumberOfHours.parse(row.find_element(By.XPATH, "td[3]").text)
            title = ActivityTitle.parse(row.find_element(By.XPATH, "td[4]").text)
            activity_type = ActivityType(row.find_element(By.XPATH, "td[5]").text)

            if with_time:
                row.find_element(By.XPATH, "td/form/input[@src = 'images/open_registro.gif']").send_keys(Keys.RETURN)
                h = self.driver.find_element(By.ID, "hh_inizio").find_element(By.XPATH, "option[@selected]").text
                m = self.driver.find_element(By.ID, "mm_inizio").find_element(By.XPATH, "option[@selected]").text
                date = date.at_time(hour=int(h), minute=int(m))
                self.driver.find_element(By.XPATH, "//a[. = 'Esci']").send_keys(Keys.RETURN)
                rows = self.driver.find_elements(By.XPATH, rows_xpath)

            res.append(RegisterActivity.of(date=date, hours=hours, activity_type=activity_type, title=title))

        return res

    def add_register_activity(self, register: Register, activity: RegisterActivity) -> bool:
        self.driver.get(URLs["register_list"])
        self.driver.find_element(
            By.XPATH,
            f"//tr[normalize-space(td/text()) = '{register.course}']/td//a[img/@src = 'images/open_registro.gif']"
        ).send_keys(Keys.RETURN)
        self.driver.find_element(
            By.XPATH,
            f"//a[normalize-space(text()) = 'Inserisci nuova attività']"
        ).send_keys(Keys.RETURN)

        self.driver.find_element(
            By.XPATH,
            "//tr[normalize-space(th/text()) = '*Data:']/td/input"
        ).send_keys(activity.date.stringify_date() + Keys.ESCAPE)
        self.driver.find_element(By.ID, "hh_inizio").send_keys(activity.date.stringify_hour())
        self.driver.find_element(By.ID, "mm_inizio").send_keys(activity.date.stringify_minute())
        self.driver.find_element(By.ID, "hh_fine").send_keys(activity.end_date_time.stringify_hour())
        self.driver.find_element(By.ID, "mm_fine").send_keys(activity.end_date_time.stringify_minute())
        self.driver.find_element(By.ID, "ore_accademiche").send_keys(str(activity.hours))
        self.driver.find_element(
            By.XPATH,
            "//tr[normalize-space(th/text()) = 'Tipo attività:']/td/select"
        ).send_keys(activity.type.value)
        self.driver.find_element(
            By.XPATH,
            "//tr[normalize-space(th/text()) = '*Titolo:']/td/input"
        ).send_keys(str(activity.title))
        self.driver.find_element(
            By.XPATH,
            "//input[@value = 'Salva']"
        ).send_keys(Keys.RETURN)

        return len(self.driver.find_elements(
            By.XPATH,
            "//tr[starts-with(normalize-space(td/text()), 'ATTENZIONE')]"
        )) == 0

    def delete_register_activity(self, register: Register, activity_index: int) -> bool:
        self.driver.get(URLs["register_list"])
        self.driver.find_element(
            By.XPATH,
            f"//tr[normalize-space(td/text()) = '{register.course}']/td//a[img/@src = 'images/open_registro.gif']"
        ).send_keys(Keys.RETURN)
        try:
            self.driver.find_element(
                By.XPATH,
                f"//tr[td/form/input/@src = 'images/open_registro.gif'][{activity_index}]/td/a[img/@src = 'images/del.gif']"
            ).send_keys(Keys.RETURN)
        except NoSuchElementException:
            return False
        self.driver.find_element(By.XPATH, "//input[@value = 'Conferma']").send_keys(Keys.RETURN)
        return True

    def fetch_graduation_days(self) -> List[GraduationDay]:
        self.driver.get(URLs["graduation_day_list"])
        rows = self.driver.find_elements(By.XPATH, "//table[@id = 'seduteAperte']/tbody/tr/td[1]")
        return [GraduationDay(row.text) for row in rows]

    def upload_graduation_day(
            self,
            graduation_day: GraduationDay,
            student_graduation_list: List[StudentGraduation],
            date: Optional[DateTime] = None,
            no_committee: bool = False,
            dry_run: bool = False,
    ) -> None:
        validate("at least one student", student_graduation_list, min_len=1, help_msg="No student was provided")
        self.driver.get(URLs["graduation_day_list"])
        self.driver.find_element(
            By.XPATH,
            f"//table[@id = 'seduteAperte']/tbody/tr/td[text() = '{graduation_day}']/../td/a"
        ).send_keys(Keys.RETURN)
        rows = self.driver.find_elements(By.XPATH, '//table[@id="elencoLaureandi"]/tbody/tr')
        student_to_url = {
            Student.of(
                student_id=row.find_element(By.XPATH, "td[2]").text,
                student_name=row.find_element(By.XPATH, "td[1]").text,
            ): row.find_element(By.XPATH, "td/a").get_attribute("href")
            for row in rows
        }
        student_to_graduation = {graduation.student: graduation for graduation in student_graduation_list}
        validate("student_graduation_list", student_to_graduation.keys(), equals=student_to_url.keys(),
                 help_msg="Students don't match")

        self.driver.get(student_to_url[student_graduation_list[0].student])
        graduation_date = self.driver.find_element(By.ID, 'grad-dettLau-dataCt').text
        if not graduation_date:
            graduation_date = DateTime.now().stringify_date() if date is None else date.stringify_date()
        if no_committee:
            committee = []
        else:
            committee = [
                checkbox.is_selected()
                for checkbox in self.driver.find_elements(
                    By.XPATH,
                    '//table[starts-with(@id, "gradDettLauCommissione")]//input[@type = "checkbox"]'
                )
            ]
            if all(not x for x in committee):
                committee = [not x for x in committee]

        for graduation in student_graduation_list:
            url = student_to_url[graduation.student]
            self.driver.get(url)

            starting_score = int(self.driver.find_element(
                By.XPATH,
                '//*[@id="grad-dettLau-boxVerbalizzazione"]/dl/dt[text() = "Voto di partenza"]/../dd'
            ).text)
            thesis_score = graduation.final_score.value - starting_score
            validate("thesis_score", thesis_score, min_value=1, max_value=15, help_msg="Wrong thesis score")

            self.__replace_content(
                self.driver.find_element(By.ID, 'grad-dettLau-annotazioni'),
                graduation.notes.value
            )
            self.__replace_content(
                self.driver.find_element(By.ID, 'grad-dettLau-puntiTesi'),
                str(thesis_score)
            )
            self.__replace_content(
                self.driver.find_element(By.ID, 'grad-dettLau-dataCt'),
                graduation_date
            )
            if graduation.laude:
                self.driver.find_element(By.ID, 'grad-dettLau-lode1').send_keys(Keys.SPACE)
            if graduation.special_mention:
                self.driver.find_element(By.ID, 'grad-dettLau-menzione1').send_keys(Keys.SPACE)

            if committee:
                checkboxes = self.driver.find_elements(
                    By.XPATH,
                    '//table[starts-with(@id, "gradDettLauCommissione")]//input[@type = "checkbox"]'
                )
                for index, member in enumerate(committee):
                    if member != checkboxes[index].is_selected():
                        checkboxes[index].send_keys(Keys.SPACE)

            if not dry_run:
                self.driver.find_element(By.ID, 'grad-dettLau-btnSubmit').send_keys(Keys.RETURN)

    @staticmethod
    def __replace_content(element, content):
        element.clear()
        element.send_keys(content + Keys.ESCAPE)



import dataclasses
from dataclasses import InitVar
from typing import Optional

import typeguard

from dumbo_esse3.utils import validators
from dumbo_esse3.utils.validators import validate


def _arithmetic(cls):
    validate(cls.__name__, hasattr(cls, '__add__'), equals=False)
    validate(cls.__name__, hasattr(cls, '__mul__'), equals=False)
    validate(cls.__name__, hasattr(cls, '__rmul__'), equals=False)
    validate(cls.__name__, hasattr(cls, '__neg__'), equals=False)
    setattr(cls, '__add__', lambda self, other: cls(self.value + (other.value if type(other) == cls else other)))
    setattr(cls, '__mul__', lambda self, other: cls(self.value * (other.value if type(other) == cls else other)))
    setattr(cls, '__rmul__', lambda self, other: cls(self.value * (other.value if type(other) == cls else other)))
    setattr(cls, '__neg__', lambda self: getattr(cls, '__mul__')(-1))


@typeguard.typechecked
def bounded_integer(min_value: int, max_value: int):
    validate('min_value', min_value, max_value=max_value)

    def decorator(cls):
        validators.does_not_has_attributes(cls, [
            '__int__',
            'min_value',
            'max_value',
            'parse',
            'of',
            '__post_init__',
            'toJSON',
        ], and_annotations=True)

        cls.__annotations__ = {'value': int}

        if getattr(cls, '__str__') == getattr(object, '__str__'):
            setattr(cls, '__str__', lambda self: str(self.value))

        setattr(cls, '__int__', lambda self: self.value)
        setattr(cls, 'min_value', staticmethod(lambda: min_value))
        setattr(cls, 'max_value', staticmethod(lambda: max_value))
        setattr(cls, 'parse', staticmethod(lambda s: cls(int(s))))
        setattr(cls, 'of', staticmethod(lambda s: cls(int(s))))

        if not hasattr(cls, '_validate'):
            setattr(cls, '_validate', lambda self: None)

        def post_init(self):
            validate('value', self.value, min_value=self.min_value(), max_value=self.max_value())
            self._validate()

        setattr(cls, '__post_init__', post_init)
        setattr(cls, 'toJSON', lambda self: self.value)
        setattr(cls, '__format__', lambda self, format_spec: self.value.__format__(format_spec))

        _arithmetic(cls)

        return dataclasses.dataclass(frozen=True, order=True)(cls)

    return decorator


@typeguard.typechecked
def bounded_string(min_length: int, max_length: int, pattern: str = r'.*', private_init: Optional[bool] = False):
    validate('min_length', min_length, min_value=0, max_value=max_length)
    validate('max_length', max_length, min_value=min_length)

    def decorator(cls):
        validators.does_not_has_attributes(cls, [
            'min_length',
            'max_length',
            'pattern',
            'parse',
            'of',
            'toJSON',
            '__post_init__',
        ], and_annotations=True)

        cls.__annotations__ = {'value': str}

        if getattr(cls, '__str__') == getattr(object, '__str__'):
            setattr(cls, '__str__', lambda self: self.value)

        setattr(cls, 'min_length', staticmethod(lambda: min_length))
        setattr(cls, 'max_length', staticmethod(lambda: max_length))
        setattr(cls, 'pattern', staticmethod(lambda: pattern))
        setattr(cls, 'parse', staticmethod(lambda s: cls(s)))
        setattr(cls, 'of', staticmethod(lambda s: cls(s)))
        setattr(cls, 'toJSON', lambda self: self.value)

        if pattern == r'.*':
            def __validate(self):
                validate('value', self.value, min_len=self.min_length(), max_len=self.max_length())
        else:
            def __validate(self):
                validate('value', self.value, min_len=self.min_length(), max_len=self.max_length(),
                         custom=validators.pattern(self.pattern()))

        if not hasattr(cls, '_validate'):
            setattr(cls, '_validate', lambda self: None)

        if private_init:
            validate('cls', hasattr(cls, '_validate_init_key'), equals=True)
            cls.__annotations__['init_key'] = InitVar

            def post_init(self, init_key):
                self._validate_init_key(init_key)
                __validate(self)
                self._validate()
        else:
            def post_init(self):
                __validate(self)
                self._validate()
        setattr(cls, '__post_init__', post_init)

        return dataclasses.dataclass(frozen=True, order=True)(cls)

    return decorator



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dumbo_esse3', 'dumbo_esse3.utils']

package_data = \
{'': ['*']}

install_requires = \
['dateutils>=0.6.12,<0.7.0',
 'rich>=12.5.1,<13.0.0',
 'selenium>=4.4.3,<5.0.0',
 'typeguard>=2.13.3,<3.0.0',
 'typer>=0.7.0,<0.8.0',
 'valid8>=5.1.2,<6.0.0']

setup_kwargs = {
    'name': 'dumbo-esse3',
    'version': '0.3.13',
    'description': 'Esse3 command line utility, to save my future time!',
    'long_description': 'None',
    'author': 'Mario Alviano',
    'author_email': 'mario.alviano@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import csv
import dataclasses
import sys
from collections import defaultdict
from pathlib import Path
from typing import Optional, List

import typer
from rich.progress import track
from rich.table import Table

from dumbo_esse3.esse3_wrapper import Esse3Wrapper
from dumbo_esse3.primitives import StudentThesisState, ExamDescription, ExamNotes, ExamType, DateTime, Register, \
    RegisterActivity, ActivityType, NumberOfHours, ActivityTitle, StudentGraduation, Student, FinalScore, GraduationNote
from dumbo_esse3.utils.console import console
from dumbo_esse3.utils.validators import validate


@dataclasses.dataclass(frozen=True)
class AppOptions:
    username: str = dataclasses.field(default='')
    password: str = dataclasses.field(default='')
    debug: bool = dataclasses.field(default=False)


app_options = AppOptions()
app = typer.Typer()


def is_debug_on():
    return app_options.debug


def run_app():
    try:
        app()
    except Exception as e:
        if is_debug_on():
            raise e
        else:
            console.print(f"[red bold]Error:[/red bold] {e}")


def new_esse3_wrapper(detached: bool = False, with_live_status: bool = True):
    def res():
        return Esse3Wrapper.create(
            username=app_options.username,
            password=app_options.password,
            debug=app_options.debug,
            detached=detached,
            headless=not app_options.debug and not detached,
        )
    if with_live_status:
        with console.status("Login..."):
            return res()
    return res()


def version_callback(value: bool):
    if value:
        import importlib.metadata
        __version__ = importlib.metadata.version("dumbo-esse3")
        console.print("dumbo-esse3", __version__)
        raise typer.Exit()


@app.callback()
def main(
        username: str = typer.Option(..., prompt=True, envvar="DUMBO_ESSE3_USERNAME"),
        password: str = typer.Option(..., prompt=True, hide_input=True, envvar="DUMBO_ESSE3_PASSWORD"),
        debug: bool = typer.Option(False, "--debug", help="Don't minimize browser"),
        version: bool = typer.Option(False, "--version", callback=version_callback, is_eager=True,
                                     help="Print version and exit"),
):
    """
    Esse3 command line utility, to save my future time!
    """
    global app_options
    app_options = AppOptions(
        username=username,
        password=password,
        debug=debug,
    )


@app.command(name="courses")
def command_courses() -> None:
    """
    Prints the list of courses.
    The number associated with each course is used an ID.
    """
    esse3_wrapper = new_esse3_wrapper()
    with console.status("Fetching courses..."):
        courses_list = esse3_wrapper.fetch_courses()

    table = Table(title="Courses")
    table.add_column("#")
    table.add_column("Course")
    for index, course in enumerate(courses_list, start=1):
        table.add_row(
            str(index),
            course.value,
        )
    console.print(table)


@app.command(name="exams")
def command_exams(
        of: List[int] = typer.Option([], help="Index of the course"),
        all_dates: Optional[bool] = typer.Option(False, "--all", "-a", help="Show all exams, not only the next"),
        with_students: Optional[bool] = typer.Option(False, "--with-students", "-s", help="Also fetch students"),
) -> None:
    """
    Prints exams and registered students.
    Filtering options are available.
    """
    esse3_wrapper = new_esse3_wrapper()
    with console.status("Fetching courses..."):
        courses = esse3_wrapper.fetch_courses()
    console.rule("Exams")
    for index, course in enumerate(track(courses, console=console, transient=True), start=1):
        if of and index not in of:
            continue
        console.print(f"{index:2d}. {course}")
        exams = esse3_wrapper.fetch_exams(course)
        if not all_dates:
            next_exam = next(filter(lambda x: x.today_or_future, exams), None)
            exams = [next_exam] if next_exam else []
        for exam in exams:
            console.print(f"  - {exam.date_and_time}, {exam.number_of_students:3d} students")
            if with_students and exam.number_of_students.positive:
                console.print()
                students = esse3_wrapper.fetch_students(course, exam.date_and_time)
                console.print("\n".join(f"{student.id}\t{student.name}" for student in students))
                console.print()


@app.command(name="add-exams")
def command_add_exams(
        exams: List[str] = typer.Argument(
            ...,
            metavar="exam",
            help="One or more strings of the form 'CourseIndex DD/MM HH:MM' "
                 "(separators can be omitted, and spaces can be replaced by dashes; year is inferred)",
        ),
        exam_type: str = typer.Option(..., prompt=True, envvar="DUMBO_ESSE3_EXAM_TYPE"),
        description: str = typer.Option(..., prompt=True, envvar="DUMBO_ESSE3_EXAM_DESCRIPTION"),
        notes: str = typer.Option(..., prompt=True, envvar="DUMBO_ESSE3_EXAM_NOTES"),
):
    """
    Adds exams provided as command-line arguments.
    """
    try:
        exam_type = ExamType(exam_type)
    except ValueError:
        console.print("Invalid type")
        raise typer.Exit()

    try:
        description = ExamDescription(description)
    except ValueError:
        console.print("Invalid description")
        raise typer.Exit()

    try:
        notes = ExamNotes(notes)
    except ValueError:
        console.print("Invalid notes")
        raise typer.Exit()

    def parse(exam):
        exam = exam.replace('-', ' ').split(' ', maxsplit=1)
        course_index = int(exam[0])
        if course_index <= 0:
            console.print(f"Course index must be positive, not {course_index}")
            raise typer.Exit()

        try:
            date = DateTime.smart_parse(exam[1])
        except ValueError:
            console.print(f"Invalid datetime. Use DD/MM HH:MM")
            raise typer.Exit()
        if date < DateTime.now():
            console.print(f"Cannot schedule an exam in the past!")
            raise typer.Exit()

        return course_index, date

    exam_list = [parse(exam) for exam in exams]

    esse3_wrapper = new_esse3_wrapper()
    courses = esse3_wrapper.fetch_courses()
    for exam in exam_list:
        if exam[0] > len(courses):
            console.print(f"Course index cannot be larger than {len(courses)}, it was given {exam[0]}")
            raise typer.Exit()

    for exam in track(exam_list, console=console, transient=True):
        index, date_and_time = exam
        course = courses[index - 1]
        if esse3_wrapper.is_exam_present(course, date_and_time):
            console.log(f"Skip already present exam {date_and_time} for {course} (#{index})")
        else:
            console.log(f"Add exam {date_and_time} to {course} (#{index})")
            esse3_wrapper.add_exam(course, date_and_time, exam_type, description, notes)


@app.command(name="theses")
def command_theses(
        list_option: bool = typer.Option(True, "--list", "--no-list", help="Print the list of theses"),
        show: List[int] = typer.Option([], help="Index of student thesis to show"),
        sign: List[int] = typer.Option([], help="Index of student thesis to sign"),
        show_all: bool = typer.Option(False, "--show-all", help="Show all theses"),
        sign_all: bool = typer.Option(False, "--sign-all", help="Sign all theses"),
):
    """
    Prints the list of theses.
    The number associated with each student is used as an ID.
    Theses can be shown in the browser and signed automatically.
    """
    esse3_wrapper = new_esse3_wrapper()
    with console.status("Fetching theses..."):
        student_thesis_states = esse3_wrapper.fetch_thesis_list()

    if list_option:
        table = Table(title="Theses")
        table.add_column("#")
        table.add_column("Student ID")
        table.add_column("Student Name")
        table.add_column("CdL")
        table.add_column("State")

        for index, student_thesis_state in enumerate(student_thesis_states, start=1):
            style = ""
            if student_thesis_state.state == StudentThesisState.State.SIGNED:
                style = "bold green"
            elif student_thesis_state.state == StudentThesisState.State.UNSIGNED:
                style = "bold red"
            table.add_row(
                str(index),
                student_thesis_state.student.id.value,
                student_thesis_state.student.name.value,
                student_thesis_state.cdl.value,
                student_thesis_state.state.name,
                style=style,
            )

        console.print(table)

    must_wait = False
    for index, student_thesis_state in enumerate(
            track(student_thesis_states, console=console, transient=True),
            start=1
    ):
        if show_all or index in show:
            if student_thesis_state.state == StudentThesisState.State.MISSING:
                console.print(f"Skip thesis of {student_thesis_state.student.name} (#{index})")
            else:
                console.print(f"Open thesis of {student_thesis_state.student.name} (#{index})")
                student_wrapper = new_esse3_wrapper(detached=True, with_live_status=False)
                student_wrapper.show_thesis(student_thesis_state.student)
                must_wait = True
    if must_wait:
        console.input("Press ENTER to continue")

    for index, student_thesis_state in enumerate(
            track(student_thesis_states, console=console, transient=True),
            start=1
    ):
        if sign_all or index in sign:
            if student_thesis_state.state == StudentThesisState.State.UNSIGNED:
                console.log(f"Sign thesis of {student_thesis_state.student.name} (#{index})")
                esse3_wrapper.sign_thesis(student_thesis_state.student)
            else:
                console.log(f"Skip thesis of {student_thesis_state.student.name} (#{index})")


@app.command(name="registers")
def command_registers() -> None:
    """
    Prints the list of registers.
    The number associated with each course is used an ID.
    """
    esse3_wrapper = new_esse3_wrapper()
    with console.status("Fetching registers..."):
        registers = esse3_wrapper.fetch_registers()

    table = Table(title="Registers")
    table.add_column("#")
    table.add_column("Course")
    table.add_column("Hours")
    table.add_column("Semester")
    table.add_column("State")
    for index, register in enumerate(registers, start=1):
        style = ""
        if register.state == Register.State.VERIFIED:
            style = "bold green"
        table.add_row(
            str(index),
            register.course.value,
            str(register.hours),
            register.semester.value,
            register.state.name if register.state == Register.State.SIGNED else
            f"[bold red]{register.state.name}[/bold red]",
            style=style,
        )
    console.print(table)


@app.command(name="register-activities")
def command_register_activities(
        of: List[int] = typer.Option([], help="Index of the register to fetch (omit to fetch all)"),
        with_time: Optional[bool] = typer.Option(False, help="Also fetch starting time of each activity"),
) -> None:
    """
    Prints register activities.
    Filtering options are available.
    """
    esse3_wrapper = new_esse3_wrapper()
    with console.status("Fetching registers..."):
        registers = esse3_wrapper.fetch_registers()

    for index, register in enumerate(track(registers, console=console, transient=True), start=1):
        if of and index not in of:
            continue
        activities = esse3_wrapper.fetch_register_activities(register, with_time)

        table = Table(title=f"{register.course}" + (
            "" if register.state == Register.State.SIGNED else f" - [red]NOT SIGNED[/red]"
        ))
        table.add_column("#")
        table.add_column("Date and time" if with_time else "Date")
        table.add_column("Hours", justify="right")
        table.add_column("Title")
        table.add_column("Type")
        for activity_index, activity in enumerate(activities, start=1):
            table.add_row(
                str(activity_index),
                str(activity.date) if with_time else activity.date.stringify_date(),
                str(activity.hours),
                str(activity.title),
                activity.type.name,
            )
        console.print(table)

        hours = defaultdict(lambda: 0)
        for activity in activities:
            hours[activity.type] += activity.hours.value
        table = Table()
        table.add_column(justify="right")
        table.add_column("Summary")
        table.add_row("State", register.state.name)
        table.add_row("Semester", str(register.semester))
        table.add_row("Hours", str(register.hours))
        table.add_row("Missing hours", str(register.hours.value - sum(activity.hours.value for activity in activities)))
        for activity_type, hours in hours.items():
            table.add_row(activity_type.name, str(hours))
        console.print(table)


@app.command(name="add-register-activity")
def command_add_register_activity(
        of: int = typer.Option(..., help="Index of the register of interest"),
        date: Optional[str] = typer.Option(DateTime.now().stringify_date(),
                                           help="Date of the activity in the format DD/MM/YYYY (today if omitted)"),
        time: str = typer.Option(..., help="Starting time of the activity in the format HH:MM"),
        hours: int = typer.Option(..., help="Number of hours of the activity"),
        title: str = typer.Option(..., help="Title of the activity"),
        activity_type: ActivityType = typer.Option(ActivityType.LECTURE.value, "--type", help="Type of activity"),
) -> None:
    """
    Add one activity to the specified register.
    """
    try:
        date_and_time = DateTime.parse(f"{date} {time}")
    except ValueError:
        console.print("Invalid date or time")
        raise typer.Exit()

    try:
        number_of_hours = NumberOfHours.of(hours)
    except ValueError:
        console.print("Invalid number of hours")
        raise typer.Exit()

    try:
        activity_title = ActivityTitle.parse(title)
    except ValueError:
        console.print("Invalid activity title")
        raise typer.Exit()

    esse3_wrapper = new_esse3_wrapper()
    with console.status("Fetching registers..."):
        registers = esse3_wrapper.fetch_registers()

    if of <= 0 or of > len(registers):
        console.print(f"Invalid register index. Must be between 1 and {len(registers)}")
        raise typer.Exit()

    register = registers[of - 1]
    activity = RegisterActivity.of(
        date=date_and_time,
        hours=number_of_hours,
        activity_type=activity_type,
        title=activity_title,
    )
    with console.status(f"Adding activity to {register.course}..."):
        added = esse3_wrapper.add_register_activity(register, activity)
        if added:
            console.log(f"Added activity to {register.course}")
        else:
            console.log(f"Cannot add activity to {register.course}. Check date and time, and try again!",
                        style="bold red")


@app.command(name="delete-register-activity")
def command_delete_register_activity(
        of: int = typer.Option(..., help="Index of the register of interest"),
        index: int = typer.Argument(..., help="Index of the activity to delete"),
) -> None:
    """
    Delete one activity of the specified register.
    """
    if index <= 0:
        console.print("Invalid index. Must be greater than 1")
        raise typer.Exit()

    esse3_wrapper = new_esse3_wrapper()
    with console.status("Fetching registers..."):
        registers = esse3_wrapper.fetch_registers()

    if of <= 0 or of > len(registers):
        console.print(f"Invalid register index. Must be between 1 and {len(registers)}")
        raise typer.Exit()

    register = registers[of - 1]
    with console.status(f"Deleting activity #{index} of register {register.course}..."):
        deleted = esse3_wrapper.delete_register_activity(register, index)
        if deleted:
            console.log(f"Deleted activity #{index} of register {register.course}")
        else:
            console.log(f"Cannot delete activity #{index} of register {register.course}. Check the index, and try again!",
                        style="bold red")


@app.command(name="graduation-days")
def command_graduation_days() -> None:
    """
    Print the list of graduation days.
    The number associated with each graduation day is used an ID.
    """
    esse3_wrapper = new_esse3_wrapper()
    with console.status("Fetching graduation days..."):
        days = esse3_wrapper.fetch_graduation_days()

    table = Table(title="Graduation days")
    table.add_column("#")
    table.add_column("Graduation day")
    for index, day in enumerate(days, start=1):
        table.add_row(
            str(index),
            day.value,
        )
    console.print(table)


def read_graduation_day_list(csv_file: Path) -> List[StudentGraduation]:
    validate("csv-file", csv_file.exists(), help_msg="The provided file doesn't exist")
    with open(csv_file) as f:
        reader = csv.reader(f)
        rows = [row for row in reader]
        student_id_index = rows[0].index("MATRICOLA")
        student_name_index = rows[0].index("STUDENTE")
        final_score_index = rows[0].index("VOTO FINALE")
        laude_index = rows[0].index("LODE")
        special_mention_index = rows[0].index("MENZIONE")
        notes_index = rows[0].index("NOTE")
        return [
            StudentGraduation(
                student=Student.of(row[student_id_index], row[student_name_index]),
                final_score=FinalScore.parse(row[final_score_index]),
                laude=bool(row[laude_index]),
                special_mention=bool(row[special_mention_index]),
                notes=GraduationNote(row[notes_index]),
            )
            for row in rows[1:]
        ]

@app.command(name="upload-graduation-day")
def command_upload_graduation_day(
        of: int = typer.Option(..., help="Index of the graduation day to upload"),
        csv_file: Path = typer.Option(..., help="Path to the CSV file containing the scores to upload"),
        date: Optional[str] = typer.Option(
            None,
            help="Date of the graduation day DD/MM/YYYY (today or from ESSE3 if omitted)"
        ),
        no_committee: bool = typer.Option(False, "--no-committee", help="Don't change selected committee"),
        dry_run: bool = typer.Option(False, "--dry-run", help="Don't save data"),
) -> None:
    """
    Upload scores for a graduation day.
    The ID of the graduation day can be obtained with the command graduation-days.
    Scores are provided via a CSV file.
    """
    student_graduation_list = read_graduation_day_list(csv_file)

    esse3_wrapper = new_esse3_wrapper()
    with console.status("Fetching graduation days..."):
        days = esse3_wrapper.fetch_graduation_days()
    validate("", of, min_value=1, max_value=len(days))

    with console.status("Fetching graduation days..."):
        esse3_wrapper.upload_graduation_day(
            graduation_day=days[of - 1],
            student_graduation_list=student_graduation_list,
            date=None if date is None else DateTime.parse_date(date),
            no_committee=no_committee,
            dry_run=dry_run,
        )
    console.print("All done!")
